/*
 * Wireless Robot using the ESP8266 in the NodeMCU Amica Module
 *
 * Robot consists of the following hardware:
 *   NodeMCU Amica module
 *   L293D Motor Controller Chip
 *
 * L293D H-Bridge Motor Controller Chip Truth Table
 *   1A/3A     2A/4A     Action
 *   ----------------------------
 *     0         0        stop
 *     1         1        stop
 *     0         1       forward
 *     1         0       reverse
 *
 * The robot controller sends commands to the server in this code.
 * Commands are: ping, lm:speed, rm:speed
 *   ping - is used for robot / robot controller synchronization
 *   lm:speed - controls the direction and speed of the left motor
 *   rm:speed - controls the direction and speed of the right motor
 *
 * Concept, Design and Implementation by: Craig A. Lindley
 * Version: 1.0
 * Last Update: 02/01/2016
 */

#include <Arduino.h>
#include <ESP8266WiFi.h>

/***********************************************************/
/***                User Configuration Items             ***/
/***********************************************************/

// Credentials for Robot's AP network
const char *ssid = "RobotNet";
const char *password = "irobot";  // Currently unused

WiFiServer server(8000);  // Server listens on port 8000

/***********************************************************/
/***             Hardware Configuration Items            ***/
/***********************************************************/

// NodeMCU GPIO pin connections to L293D motor controller
#define LM_DIR_PIN    4
#define LM_SPEED_PIN  0
#define RM_DIR_PIN   15
#define RM_SPEED_PIN  2

// Motor identifiers
#define LEFT_MOTOR    0
#define RIGHT_MOTOR   1

// Motor direction identifiers
#define FORWARD       0
#define BACKWARD      1

/***********************************************************/
/***                   Program Utilities                 ***/
/***********************************************************/

// Motor control function
void setMotor(int motorID, int dir, int speed) {
  if (motorID == LEFT_MOTOR) {
    if (dir == FORWARD) {
      // We are going forward
      digitalWrite(LM_DIR_PIN, 0);
      analogWrite(LM_SPEED_PIN, speed);
    } else  {
      // We are going backward
      digitalWrite(LM_DIR_PIN, 1);
      analogWrite(LM_SPEED_PIN, 1023 - speed);
    }
  } else  {
    if (dir == FORWARD) {
      // We are going forward
      digitalWrite(RM_DIR_PIN, 0);
      analogWrite(RM_SPEED_PIN, speed);
    } else  {
      // We are going backward
      digitalWrite(RM_DIR_PIN, 1);
      analogWrite(RM_SPEED_PIN, 1023 - speed);
    }
  }
  delay(25);
}

/***********************************************************/
/***                    Program Setup                    ***/
/***********************************************************/

void setup() {

  Serial.begin(115200);
  delay(2000);

  // Set GPIO pins for the motor controller chip to outputs
  pinMode(LM_DIR_PIN,   OUTPUT);
  pinMode(LM_SPEED_PIN, OUTPUT);
  pinMode(RM_DIR_PIN,   OUTPUT);
  pinMode(RM_SPEED_PIN, OUTPUT);

  // Set GPIO pins LOW to stop the motors
  digitalWrite(LM_DIR_PIN,   LOW);
  digitalWrite(LM_SPEED_PIN, LOW);
  digitalWrite(RM_DIR_PIN,   LOW);
  digitalWrite(RM_SPEED_PIN, LOW);

  Serial.println("Configuring access point...");
  // WiFi.softAP(ssid, password);
  WiFi.softAP(ssid);

  IPAddress myIP = WiFi.softAPIP();
  Serial.print("Access Point IP address: ");
  Serial.println(myIP);
  Serial.flush();

  // Enable WiFi server
  server.begin();
  server.setNoDelay(true);
}

/***********************************************************/
/***                    Main Program                     ***/
/***********************************************************/

WiFiClient client;

void loop() {

  // Has a client attached to this server
  client = server.available();
  if (! client) {
    return;
  }
  Serial.println("Got connection");

  while (client.connected()) {
    while (client.available()) {
      // Process incoming data
      String line = client.readStringUntil('\n');

      // Remove any surrounding whitespace chars
      line.trim();

      // Did we receive a ping from the controller ?
      if (line.equals("ping")) {
        // Send the ping back to the controller
        client.println(line);
        continue;
      }

      // Did we receive a left motor command ?
      if (line.startsWith("lm:")) {
        String lmStr = line.substring(3);
        int lmSpeed = lmStr.toInt();
        if (lmSpeed >= 0) {
          setMotor(LEFT_MOTOR, FORWARD, abs(lmSpeed));
        } else  {
          setMotor(LEFT_MOTOR, BACKWARD, abs(lmSpeed));
        }
        continue;
      }

      // Did we receive a right motor command ?
      if (line.startsWith("rm:")) {
        String rmStr = line.substring(3);
        int rmSpeed = rmStr.toInt();
        if (rmSpeed >= 0) {
          setMotor(RIGHT_MOTOR, FORWARD, abs(rmSpeed));
        } else  {
          setMotor(RIGHT_MOTOR, BACKWARD, abs(rmSpeed));
        }
        continue;
      }
    }
    delay(1);
  }
  client.stop();
}
