/*
 * Wireless Robot Remote Control using the ESP8266 in the NodeMCU Amica Module
 *
 * Controller consists of the following hardware:
 *   NodeMCU Amica module
 *   Joystick with switch
 *   CD4066 analog switch for expanding ADC to two instead of one input
 *   RGB LED (common cathode)
 *
 * Concept, Design and Implementation by: Craig A. Lindley
 * Version: 1.1
 * Last Update: 02/15/2016
 */

#include <Arduino.h>
#include <ESP8266WiFi.h>

/***********************************************************/
/***                User Configuration Items             ***/
/***********************************************************/

// WiFi login credentials
const char *WIFI_SSID = "RobotNet";
const char *WIFI_PASS = "irobot"; // Currently unused

// Access Point Server's IP address and port
const char *SERVER_IP = "192.168.4.1";
const int SERVER_PORT = 8000;

// Sets the maximum speed of the robot's motors
const int MAX_SPEED = 900;

/***********************************************************/
/***             Hardware Configuration Items            ***/
/***********************************************************/

#define JS_SW          0  // Joystick switch

#define JS_X_SELECTOR  5  // Analog switch selector for X joystick
#define JS_Y_SELECTOR  4  // Analog switch selector for Y joystick

#define LED_RED       14  // Red portion of RGB LED
#define LED_GRN       12  // Green portion of RGB LED
#define LED_BLU       13  // Blue portion of RGB LED

// Color definitions
#define COLOR_BLK 0x00000000
#define COLOR_RED 0x00FF0000
#define COLOR_GRN 0x0000FF00
#define COLOR_BLU 0x000000FF
#define COLOR_YEL 0x00FFFF00
#define COLOR_MAG 0x00FF00FF
#define COLOR_CYN 0x0000FFFF
#define COLOR_WHT 0x00FFFFFF

// Joystick definitions
// X joystick moves left to right
// Y joystick moves forward and back
#define X_JS  0
#define Y_JS  1

// Deadzone around center joystick position
#define DEADZONE 8

// Finite State Machine (FSM) States
enum STATES {
  STATE_INIT,
  STATE_CHECK_CONNECTION,
  STATE_MAKE_CONNECTION,
  STATE_CHECK_PINGS,
  STATE_CHECK_INCOMING,
  STATE_PROCESS_JOYSTICK,
  STATE_ERROR,
};

// Current state the FSM is in
enum STATES state;

// WiFi client
WiFiClient client;

// Joystick centered values
int JS_MID_X;
int JS_MID_Y;

/***********************************************************/
/***                   Program Utilities                 ***/
/***********************************************************/

// printf equivalent
void p(const char *fmt, ... ) {
  char buf[50]; // resulting string limited to 50 chars
  va_list args;
  va_start (args, fmt );
  vsnprintf(buf, 50, fmt, args);
  va_end (args);
  Serial.print(buf);
}

// Read the specified joystick's position
// Returns value between 0 .. 1023
int readJoystickPosition(int selector) {

  if (selector == X_JS) {
    // Prepare to read X joystick
    // Route X JS to analog input
    digitalWrite(JS_Y_SELECTOR, LOW);
    digitalWrite(JS_X_SELECTOR, HIGH);

  } else  {

    // Prepare to read Y joystick
    // Route Y JS to analog input
    digitalWrite(JS_X_SELECTOR, LOW);
    digitalWrite(JS_Y_SELECTOR, HIGH);
  }
  // Delay to let level settle
  delay(5);

  // Read and return joystick's value
  return analogRead(A0);
}

// Read the joystick's switch
// Returns 1 if switch is pushed; 0 otherwise
int readJoystickSwitch() {
  return ! digitalRead(JS_SW);
}

// Wait for joystick switch press
void waitForJoystickSwitch() {
  while (digitalRead(JS_SW) == HIGH) {
    // Delay so ESP8266 watchdog timer won't trip
    delay(1);
  }
}

// Set the RGB LED to the specified 24 bit color
// using 3 channel PWM.
// Color encoded as 0x00RRGGBB
void setLEDColor(uint32_t color) {
  int red, grn, blu;

  red = (color >> 16) & 0xff;
  grn = (color >>  8) & 0xff;
  blu = (color)       & 0xff;

  analogWrite(LED_RED, red << 2);
  analogWrite(LED_GRN, grn << 2);
  analogWrite(LED_BLU, blu << 2);
}

/***********************************************************/
/***                    Program Setup                    ***/
/***********************************************************/

void setup() {

  Serial.begin(115200);
  delay(2000);

  // Configure the I/O lines
  // Assign input with pullup for the joystick switch
  pinMode(JS_SW, INPUT_PULLUP);

  // Assign LED lines as output
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GRN, OUTPUT);
  pinMode(LED_BLU, OUTPUT);

  // Turn off RGB LED for now
  digitalWrite(LED_RED, LOW);
  digitalWrite(LED_GRN, LOW);
  digitalWrite(LED_BLU, LOW);

  // Output pins for joystick analog switches
  pinMode(JS_X_SELECTOR, OUTPUT);
  pinMode(JS_Y_SELECTOR, OUTPUT);

  // Turn off joystick's analog switches for now
  digitalWrite(JS_X_SELECTOR, LOW);
  digitalWrite(JS_Y_SELECTOR, LOW);

  // Record the center position values of the joystick
  JS_MID_X = readJoystickPosition(X_JS);
  JS_MID_Y = readJoystickPosition(Y_JS);

  p("\r\nRobot Controller\r\n");

  // Connect to open WiFi access point network
  WiFi.mode(WIFI_STA);  // Must set this or will send packets to itself
  WiFi.begin(WIFI_SSID);

  uint8_t i = 0;
  while ((WiFi.status() != WL_CONNECTED) && (i++ < 30)) {
    delay(500);
    p(".");
  }
  if (i == 31) {
    p("\r\nCould not connect to: %s\n", WIFI_SSID);
    // Blink LED red to indicate no network connection
    while (true) {
      setLEDColor(COLOR_RED);
      delay(500);
      setLEDColor(COLOR_BLK);
      delay(500);
    }
  }
  p("\r\nConnected to: %s\n", WIFI_SSID);
  delay(500);

  // Set initial state machine state
  state = STATE_INIT;
}

/***********************************************************/
/***                    Program Loop                     ***/
/***********************************************************/
#define SEND_PING_TIMEOUT_MS 1000
#define RECV_PING_TIMEOUT_MS 4000
#define ERROR_BLINK_COUNT 5

void loop() {
  int x, y;
  static char buffer[20];
  static uint32_t errorColor, sendPingTimeout, recvPingTimeout;
  static int oldLeft, oldRight;

  delay(5);

  // Run state machine
  switch (state) {

    case STATE_INIT:
      // Initialize variables
      errorColor = COLOR_BLK;
      oldLeft = oldRight = 0;
      sendPingTimeout = 0;
      recvPingTimeout = millis() + RECV_PING_TIMEOUT_MS;

      // Set LED to red until connection is made
      setLEDColor(COLOR_RED);

      // Set next state
      state = STATE_CHECK_CONNECTION;
      break;

    case STATE_CHECK_CONNECTION:
      if (! client.connected()) {
        state = STATE_MAKE_CONNECTION;
      } else  {
        state = STATE_CHECK_PINGS;
      }
      break;

    case STATE_MAKE_CONNECTION:
      // Attempt to connect to robot
      if (! client.connect(SERVER_IP, SERVER_PORT)) {
        errorColor = COLOR_RED;
        state = STATE_ERROR;
      } else  {
        // Indicate connection made
        setLEDColor(COLOR_GRN);
        state = STATE_CHECK_PINGS;
      }
      break;

    case STATE_CHECK_PINGS:
      // Is it time to send a ping to the robot ?
      if (millis() > sendPingTimeout) {

        // Ping the robot so we know we are still in control
        client.println("ping");
        delay(20);

        // Reset timer
        sendPingTimeout = millis() + SEND_PING_TIMEOUT_MS;
        break;
      }

      // Have we missed a received ping ?
      if (millis() > recvPingTimeout) {
        errorColor = COLOR_BLU;
        state = STATE_ERROR;
        break;
      }
      state = STATE_CHECK_INCOMING;
      break;

    case STATE_CHECK_INCOMING:
      // Is there incoming data from the robot ?
      while (client.available()) {
        // Yes there is some data
        String line = client.readStringUntil('\n');

        // Remove any surrounding whitespace chars
        line.trim();

        //        p("Rcv: %s\n", line.c_str());

        // Is it the ping we're expecting ?
        if (line.equals("ping")) {
          // Yes, reset timeout
          recvPingTimeout = millis() + RECV_PING_TIMEOUT_MS;
        }
        delay(1);
      }
      state = STATE_PROCESS_JOYSTICK;
      break;

    case STATE_PROCESS_JOYSTICK:
      {
        // Read the joystick's position
        int xRaw = readJoystickPosition(X_JS);
        int yRaw = readJoystickPosition(Y_JS);

        // Create a deadzone around the middle joystick position
        // If value within deadzone, set value to center value
        if (abs(xRaw - JS_MID_X) < DEADZONE) {
          xRaw = JS_MID_X;
        }
        if (abs(yRaw - JS_MID_Y) < DEADZONE) {
          yRaw = JS_MID_Y;
        }

        // Determine which quadrant the joystick is in
        bool xNeg = xRaw < JS_MID_X;
        bool yNeg = yRaw > JS_MID_Y;

        // Get the absolute value of the joystick's displacement from center
        int x = abs(xRaw - JS_MID_X);
        int y = abs(yRaw - JS_MID_Y);

        // Clamp upper values to max value of 500
        x = (x > 500) ? 500 : x;
        y = (y > 500) ? 500 : y;

        // Speed is the the larger of the displacement values
        int speed = (x >= y) ? x : y;

        // Initially set both motors to the same speed
        int left  = speed;
        int right = speed;

        if (xNeg) {
          // If the joystick is to the left of center
          // slow down the left motor so robot will turn left
          left -= x/2;
        } else  {
          // If the joystick is to the right of center
          // slow down the right motor so robot will turn right
          right -= x/2;
        }

        // Map the calculated speeds into motor range 0 .. MAX_SPEED
        left  = map(left,  0, 500, 0, MAX_SPEED);
        right = map(right, 0, 500, 0, MAX_SPEED);

        // If joystick was below center, reverse both motors
        if (yNeg) {
          left  *= -1;
          right *= -1;
        }

        yield();  // Just in case

        // Only send new messages to robot if joystick's position has changed
        if ((oldLeft != left) || (oldRight != right)) {
          // Display the left and right motor values
          p("left: %d, right: %d\n", left, right);

          // Format and print messages to robot
          sprintf(buffer, "lm:%d", left);
          client.println(buffer);
          delay(20);
          sprintf(buffer, "rm:%d", right);
          client.println(buffer);
          delay(20);

          // Save old values for next time
          oldLeft = left;
          oldRight = right;
        }
        state = STATE_CHECK_CONNECTION;
      }
      break;

    case STATE_ERROR:
      // Force server to disconnect
      client.stop();

      // Blink the error indicator LED
      for (int i = 0; i < ERROR_BLINK_COUNT; i++) {
        setLEDColor(errorColor);
        delay(500);
        setLEDColor(COLOR_BLK);
        delay(500);
      }
      // Attempt to restart
      state = STATE_INIT;
      delay(1000);
      break;
  }
}
